-- ===========================================================================
--	World Icon Manager
--	Handles showing world icons (e.g., resources) on the map.
--	May be moved to PlotInfo.lua
-- ===========================================================================

include( "InstanceManager" );
include( "SupportFunctions" );

local KEY_CURRENT_ICON_INFO		:string = "currentIcon";
local KEY_PREVIOUS_ICON_INFO	:string = "prevIcon";
local KEY_PLOT_INDEX			:string = "plotIndex";

local g_InstanceManager			:table = InstanceManager:new( "IconSetInstance",	"Anchor", Controls.IconContainer );
local g_MapIcons				:table = {};
local m_kUntouchedPlots			:table = {};	-- Used to prevent multiple calls to mess with animation start for a plot change
local m_isShowResources			:boolean = UserConfiguration.ShowMapResources();
local m_isShowRecommendations	:boolean = true;

local m_techsThatUnlockResources : table = {};
local m_civicsThatUnlockResources : table = {};
local m_techsThatUnlockImprovements : table = {};

-- Stores a list of plot indexes currently showing improvement recommendations
-- Used to efficiently clear improvement recommendations
local m_RecommendedImprovementPlots	:table = {};

-- Stores a list of plot indexes currently showing settlement recommendations
-- Used to efficiently clear settlement recommendations
local m_RecommendedSettlementPlots	:table = {};


-- RESOURCE TOOLTIP MOD
-- Stores all resources and the amount the player currently owns
local g_ResourcesForPlayer :table = {};
local g_ResourcesOnMap     :table = {};
-- END RESOURCE TOOLTIP MOD


-- ===========================================================================
function GetStartingPlotPlayer( pPlot )
	local x = pPlot:GetX();
	local y = pPlot:GetY();

	for i = 0, GameDefines.MAX_PLAYERS-1 do
		local playerConfig = PlayerConfigurations[i];
		if (playerConfig:IsInUse()) then
			local location = playerConfig:GetStartingPosition();
			if (location.x == x and location.y == y) then
				return i;
			end
		end
	end

	return -1;
end

-- ===========================================================================
--	Animation Callback
--	The new icon state is done fading in, set the texture it on the 
--	non-animating control and clear out the animating control.
-- ===========================================================================
function OnEndFade( pInstance:table )
	local pCurrentIconInfo = pInstance[KEY_CURRENT_ICON_INFO];
	if (pCurrentIconInfo ~= nil) then
		local textureOffsetX:number = pCurrentIconInfo.textureOffsetX;
		local textureOffsetY:number = pCurrentIconInfo.textureOffsetY;
		local textureSheet:string	= pCurrentIconInfo.textureSheet;
		pInstance.ResourceIcon:SetTexture( textureOffsetX, textureOffsetY, textureSheet );
		pInstance.ResourceIcon:SetHide( not m_isShowResources );
		--pInstance.NextResourceIcon:UnloadTexture();
		pInstance.NextResourceIcon:SetHide( true );
		m_kUntouchedPlots[pInstance[KEY_PLOT_INDEX]] = nil;
	end
end

-------------------------------------------------------------------------------
function SetResourceIcon( pInstance:table, pPlot, type, state)
	local resourceInfo = GameInfo.Resources[type];
	if (pPlot and resourceInfo ~= nil) then
		local resourceType:string = resourceInfo.ResourceType;
		local featureType :string;
		local terrainType :string;

		local feature = GameInfo.Features[pPlot:GetFeatureType()];
		if(feature) then
			featureType = feature.FeatureType;
		end

		local terrain = GameInfo.Terrains[pPlot:GetTerrainType()];
		if(terrain) then
			terrainType = terrain.TerrainType;
		end

		local iconName = "ICON_" .. resourceType;
		if (state == RevealedState.REVEALED) then
			iconName = iconName .. "_FOW";
		end
		local textureOffsetX, textureOffsetY, textureSheet = IconManager:FindIconAtlas(iconName, 64);
		if (textureSheet ~= nil) then						
			pInstance[KEY_PREVIOUS_ICON_INFO] = DeepCopy( pInstance[KEY_CURRENT_ICON_INFO] );
			if pInstance[KEY_PREVIOUS_ICON_INFO] ~= nil then
				pInstance.ResourceIcon:SetTexture( pInstance[KEY_PREVIOUS_ICON_INFO].textureOffsetX, pInstance[KEY_PREVIOUS_ICON_INFO].textureOffsetY, pInstance[KEY_PREVIOUS_ICON_INFO].textureSheet );
				pInstance.ResourceIcon:SetHide( not m_isShowResources );
			else
				pInstance.ResourceIcon:SetHide( true );
			end			
			pInstance.NextResourceIcon:SetTexture( textureOffsetX, textureOffsetY, textureSheet );
			pInstance.NextResourceIcon:SetHide( false );
			pInstance[KEY_CURRENT_ICON_INFO] = {
				textureOffsetX = textureOffsetX, 
				textureOffsetY = textureOffsetY,
				textureSheet  = textureSheet
			}
			pInstance.AlphaAnim:SetHide(false);
			pInstance.AlphaAnim:SetToBeginning();
			pInstance.AlphaAnim:Play();
			
			-- Add some tooltip information about the resource
			local toolTipItems:table = {};
			table.insert(toolTipItems, Locale.Lookup(resourceInfo.Name));
			if (resourceInfo.ResourceClassType == "RESOURCECLASS_BONUS") then
				table.insert(toolTipItems, Locale.Lookup("LOC_TOOLTIP_BONUS_RESOURCE"));
			elseif (resourceInfo.ResourceClassType == "RESOURCECLASS_LUXURY") then
				table.insert(toolTipItems, Locale.Lookup("LOC_TOOLTIP_LUXURY_RESOURCE"));
			elseif (resourceInfo.ResourceClassType == "RESOURCECLASS_STRATEGIC") then
				table.insert(toolTipItems, Locale.Lookup("LOC_TOOLTIP_STRATEGIC_RESOURCE"));
			elseif (resourceInfo.ResourceClassType == "RESOURCECLASS_ARTIFACT") then
				table.insert(toolTipItems, Locale.Lookup("LOC_TOOLTIP_ARTIFACT_RESOURCE"));
				table.insert(toolTipItems, Locale.Lookup("LOC_TOOLTIP_ARTIFACT_RESOURCE_DETAILS"));
			end
			local resourceTechType;

			local tValidImprovements:table = {}
			for row in GameInfo.Improvement_ValidResources() do
				if (row.ResourceType == resourceType) then
					table.insert(tValidImprovements, row.ImprovementType);
				end
			end

			local resourceTechType;
			if (table.count(tValidImprovements) > 0) then
				if (table.count(tValidImprovements) > 1) then
					for i, improvement in ipairs(tValidImprovements) do
						local improvementType = improvement;

						local has_feature = false;
						local valid_feature = false;
						for inner_row in GameInfo.Improvement_ValidFeatures() do
							if(inner_row.ImprovementType == improvementType) then
								has_feature = true;
								if(inner_row.FeatureType == featureType) then
									valid_feature = true;
								end
							end
						end
						valid_feature = not has_feature or valid_feature;

						local has_terrain = false;
						local valid_terrain = false;
						for inner_row in GameInfo.Improvement_ValidTerrains() do
							if(inner_row.ImprovementType == improvementType) then
								has_terrain = true;
								if(inner_row.TerrainType == terrainType) then
									valid_terrain = true;
								end
							end
						end
						valid_terrain = not has_terrain or valid_terrain;

						if(valid_feature and valid_terrain) then
							resourceTechType = GameInfo.Improvements[improvementType].PrereqTech;
							break;
						end
					end
				else
					local improvementType = tValidImprovements[1];
					resourceTechType = GameInfo.Improvements[improvementType].PrereqTech;
				end
			end

			if (resourceTechType ~= nil) then
				local localPlayer	= Players[Game.GetLocalPlayer()];
				if (localPlayer ~= nil) then
					local playerTechs	= localPlayer:GetTechs();
					local techType = GameInfo.Technologies[resourceTechType];
					if (techType ~= nil and not playerTechs:HasTech(techType.Index)) then
						table.insert(toolTipItems,"[COLOR:Civ6Red](".. Locale.Lookup("LOC_TOOLTIP_REQUIRES") .. " " .. Locale.Lookup(techType.Name) .. ")[ENDCOLOR]");
					end
				end
			end


			-- RESOURCE TOOLTIP MOD
			-- Add the amount of resources already available to you
			local playerID:number  = Game.GetLocalPlayer();
			local player:table     = Players[playerID];
			local pResources:table = player:GetResources();

			if ( pResources ) then
				for row in GameInfo.Resources() do
					if ( resourceInfo.Index == row.Index ) then
						toolTipItems = createResourceTooltipString(toolTipItems, pResources, row);
					end
				end
			end
			-- END RESOURCE TOOLTIP MOD

			pInstance.ResourceIcon:SetToolTipString(table.concat(toolTipItems, "[NEWLINE]"));
		end
	end
end

-------------------------------------------------------------------------------
function UnloadResourceIconAt(plotIndex)
	local pInstance = g_MapIcons[plotIndex];
	if (pInstance ~= nil) then
		pInstance.NextResourceIcon:UnloadTexture();
		pInstance.NextResourceIcon:SetHide( true );
		pInstance.AlphaAnim:SetHide( true );
		pInstance.ResourceIcon:UnloadTexture();
		pInstance.ResourceIcon:SetHide( true );
	end
end

-------------------------------------------------------------------------------
function UnloadRecommendationIconAt(plotIndex)
	local pInstance = g_MapIcons[plotIndex];
	if (pInstance ~= nil) then
		pInstance.RecommendationIconTexture:UnloadTexture();
		pInstance.RecommendationIcon:SetHide( true );
	end
end

-------------------------------------------------------------------------------
function GetInstanceAt(plotIndex)
	local pInstance = g_MapIcons[plotIndex];
	if (pInstance == nil) then
		pInstance = g_InstanceManager:GetInstance();
		g_MapIcons[plotIndex] = pInstance;
		local worldX, worldY = UI.GridToWorld( plotIndex );
		pInstance.Anchor:SetWorldPositionVal( worldX, worldY-10.0, 0.0 );
		-- Do not unload the texture on the ResourceIcon itself, it may remove the only instance to be animated in.
		pInstance.ResourceIcon:SetHide( true );
		-- Do not show/start the AlphaAnim, it should not run if there is only the recommendation icon in the hex
		pInstance.AlphaAnim:SetHide( true );
		pInstance.RecommendationIconTexture:UnloadTexture();
		pInstance.RecommendationIcon:SetHide( true );
		pInstance.AlphaAnim:RegisterEndCallback( function() OnEndFade(pInstance); end );
		pInstance[KEY_PLOT_INDEX] = plotIndex;
	end
	return pInstance;
end

-------------------------------------------------------------------------------
function GetInstanceAllocatedAt(plotIndex)
	return g_MapIcons[plotIndex];
end

-------------------------------------------------------------------------------
function ReleaseInstanceAt(plotIndex)
	local pInstance = g_MapIcons[plotIndex];
	if (pInstance ~= nil) then
		pInstance.ResourceIcon:UnloadTexture();
		pInstance.RecommendationIconTexture:UnloadTexture();
		g_InstanceManager:ReleaseInstance( pInstance );
		g_MapIcons[plotIndex] = nil;
	end
end

-------------------------------------------------------------------------------
function GetNonEmptyAt(plotIndex, state)

	local eObserverID = Game.GetLocalObserver();
	local pLocalPlayerVis = PlayerVisibilityManager.GetPlayerVisibility(eObserverID);
	if (pLocalPlayerVis ~= nil) then
		local pInstance = nil;

		local pPlot = Map.GetPlotByIndex(plotIndex);
		-- Have a Resource?
		local eResource = pLocalPlayerVis:GetLayerValue(VisibilityLayerTypes.RESOURCES, plotIndex);
		local bHideResource = ( pPlot ~= nil and ( pPlot:GetDistrictType() > 0 or pPlot:IsCity() ) );
		if (eResource ~= nil and eResource ~= -1 and not bHideResource ) then
			pInstance = GetInstanceAt(plotIndex);
			SetResourceIcon(pInstance, pPlot, eResource, state);
		else
			UnloadResourceIconAt(plotIndex);
		end

		if (pPlot) then
			-- Starting plot?
			if pPlot:IsStartingPlot() and WorldBuilder.IsActive() then
				pInstance = GetInstanceAt(plotIndex);
				pInstance.RecommendationIconTexture:SetTexture( IconManager:FindIconAtlas("ICON_UNITOPERATION_FOUND_CITY", 38) );
				pInstance.RecommendationIconText:SetHide( false );

				local iPlayer = GetStartingPlotPlayer( pPlot );
				if (iPlayer >= 0) then
					pInstance.RecommendationIconText:SetText( tostring(iPlayer + 1) );
				else
					pInstance.RecommendationIconText:SetText( "" );
				end
			else
				UnloadRecommendationIconAt(plotIndex);
			end
		end
		return pInstance;
	end
end

-------------------------------------------------------------------------------
function RemoveAll(plotIndex)
	ReleaseInstanceAt(plotIndex);
end
	
-------------------------------------------------------------------------------
function ChangeToMidFog(plotIndex)
	local pIconSet = GetNonEmptyAt(plotIndex, RevealedState.REVEALED);

	if (pIconSet ~= nil) then
		pIconSet.NextResourceIcon:SetHide( (not m_isShowResources) or (not pIconSet.ResourceIcon:HasTexture()) );
		pIconSet.RecommendationIcon:SetHide( (not m_isShowRecommendations) or (not pIconSet.RecommendationIconTexture:HasTexture()) );
		pIconSet.IconStack:CalculateSize();
		pIconSet.IconStack:ReprocessAnchoring();
	end
end

-------------------------------------------------------------------------------
function ChangeToVisible(plotIndex)
	
	local pIconSet = GetNonEmptyAt(plotIndex, RevealedState.VISIBLE);

	if (pIconSet ~= nil) then
		pIconSet.NextResourceIcon:SetHide( (not m_isShowResources) or (not pIconSet.ResourceIcon:HasTexture()) );
		pIconSet.RecommendationIcon:SetHide( (not m_isShowRecommendations) or (not pIconSet.RecommendationIconTexture:HasTexture()) );
		pIconSet.IconStack:CalculateSize();
		pIconSet.IconStack:ReprocessAnchoring();
	end
end

-------------------------------------------------------------------------------
function Rebuild()

	local eObserverID = Game.GetLocalObserver();
	local pLocalPlayerVis = PlayerVisibilityManager.GetPlayerVisibility(eObserverID);

	if (pLocalPlayerVis ~= nil) then
		local iCount = Map.GetPlotCount();
		for plotIndex = 0, iCount-1, 1 do

			local visibilityType = pLocalPlayerVis:GetState(plotIndex);
			if (visibilityType == RevealedState.HIDDEN) then
				RemoveAll(plotIndex);
			else		
				if (visibilityType == RevealedState.REVEALED) then
					ChangeToMidFog(plotIndex);
				else
					if (visibilityType == RevealedState.VISIBLE) then
						ChangeToVisible(plotIndex);
					end
				end
			end
		end
	end
end

function OnResearchCompleted( player:number, tech:number, isCanceled:boolean)
	if player == Game.GetLocalPlayer() then
		for i, techType in ipairs(m_techsThatUnlockResources) do
			if (techType == GameInfo.Technologies[tech].TechnologyType) then
				Rebuild();
			end
		end

		for i, techType in ipairs(m_techsThatUnlockImprovements) do
			if (techType == GameInfo.Technologies[tech].TechnologyType) then
				Rebuild();
				break;
			end
		end
	end
end

function OnCivicCompleted( player:number, civic:number, isCanceled:boolean)
	if player == Game.GetLocalPlayer() then
		for i, civicType in ipairs(m_civicsThatUnlockResources) do
			if (civicType == GameInfo.Civics[civic].CivicType) then
				Rebuild();
			end
		end
	end
end

-------------------------------------------------------------------------------
function OnResourceVisibilityChanged(x, y, resourceType, visibilityType)

	-- Don't use the 'untouched' version because events this can break depending on the order the events come in.
	local plotIndex:number = GetPlotIndex(x, y);
	if plotIndex == -1 then
		return;
	end

	if (visibilityType == RevealedState.HIDDEN) then
		UnloadResourceIconAt(plotIndex);
	else		
		if (visibilityType == RevealedState.REVEALED) then
			ChangeToMidFog(plotIndex);
		else
			if (visibilityType == RevealedState.VISIBLE) then
				ChangeToVisible(plotIndex);
			end
		end
	end
end

function OnResourceRemovedFromMap(x, y, resourceType )
	OnResourceVisibilityChanged(x, y, resourceType, RevealedState.HIDDEN );
end

-------------------------------------------------------------------------------
function OnResourceChanged(x, y, resourceType)
	local eObserverID = Game.GetLocalObserver();
	local pLocalPlayerVis = PlayerVisibilityManager.GetPlayerVisibility(eObserverID);
	if (pLocalPlayerVis ~= nil) then

		local visibilityType	= pLocalPlayerVis:GetState(x, y);

		-- Don't use the 'untouched' version because events this can break depending on the order the events come in.
		local plotIndex:number = GetPlotIndex(x, y);
		if plotIndex == -1 then
			return;
		end

		if (visibilityType == RevealedState.HIDDEN) then
			UnloadResourceIconAt(plotIndex);
		else		
			if (visibilityType == RevealedState.REVEALED) then
				ChangeToMidFog(plotIndex);
			else
				if (visibilityType == RevealedState.VISIBLE) then
					ChangeToVisible(plotIndex);
				end
			end
		end

	end
end

-------------------------------------------------------------------------------
function OnPlotVisibilityChanged(x, y, visibilityType)

	-- Don't use the 'untouched' version because events this can break depending on the order the events come in.
	local plotIndex:number = GetPlotIndex(x, y);
	if plotIndex == -1 then
		return;
	end

	if (visibilityType == RevealedState.HIDDEN) then
		RemoveAll(plotIndex);
	else		
		if (visibilityType == RevealedState.REVEALED) then
			ChangeToMidFog(plotIndex);
		else
			if (visibilityType == RevealedState.VISIBLE) then
				ChangeToVisible(plotIndex);
			end
		end
	end
end

-- ===========================================================================
function OnPlotMarkersChanged(x, y)

	-- The marker for a plot has changed.
	local eObserverID = Game.GetLocalObserver();
	local pLocalPlayerVis = PlayerVisibilityManager.GetPlayerVisibility(eObserverID);
	if (pLocalPlayerVis ~= nil) then
		
		local visibilityType	= pLocalPlayerVis:GetState(x, y);
		local plotIndex	:number	= GetUntouchPlotIndex(x, y);
		if plotIndex == -1 then
			return;
		end

		if (visibilityType == RevealedState.REVEALED) then
			ChangeToMidFog(plotIndex);
		else
			if (visibilityType == RevealedState.VISIBLE) then
				ChangeToVisible(plotIndex);
			end
		end
	end
end


-- ===========================================================================
--	Returns the plot # for a given set of coordinates or returns -1 if
--	it has already been handed out this turn.
-- ===========================================================================
function GetUntouchPlotIndex(x,y)
	local plotIndex	:number = Map.GetPlotIndex(x, y);
	if WorldBuilder.IsActive() then
		return plotIndex;
	else
		local gameTurn	:number	= Game.GetCurrentGameTurn();
		if m_kUntouchedPlots[plotIndex] == nil or m_kUntouchedPlots[plotIndex] ~= gameTurn then
			m_kUntouchedPlots[plotIndex] = gameTurn;
			return plotIndex;
		end

		return -1;
	end
end

-- ===========================================================================
-- Returns the plot # for a given set of coordinates.
-- ===========================================================================
function GetPlotIndex(x,y)
	local plotIndex	:number = Map.GetPlotIndex(x, y);
	return plotIndex;
end



----------------------------------------------------------------
function OnLocalPlayerChanged( eLocalPlayer:number , ePrevLocalPlayer:number )
	for key, pIconSet in pairs(g_MapIcons) do
		if (pIconSet ~= nil) then
			g_InstanceManager:ReleaseInstance( pIconSet );
			g_MapIcons[key] = nil;
		end
	end
	m_kUntouchedPlots = {};
end


----------------------------------------------------------------
function OnUserOptionChanged(eOptionSet, hOptionKey, iNewOptionValue)

	local bChangedValue = UserConfiguration.ShowMapResources();
	if (bChangedValue ~= m_isShowResources) then
		m_isShowResources = bChangedValue;

		for _, pIconSet in pairs(g_MapIcons) do
			pIconSet.ResourceIcon:SetHide( (not m_isShowResources) or (not pIconSet.ResourceIcon:HasTexture()) );
			pIconSet.IconStack:CalculateSize();
			pIconSet.IconStack:ReprocessAnchoring();
		end
	end
end
----------------------------------------------------------------
-- Handle the UI shutting down.
function OnShutdown()
	g_MapIcons = {};
	g_InstanceManager:DestroyInstances();
end

-- ===========================================================================
function OnContextInitialize(bHotload : boolean)
	-- The context just loaded, is it a hotload?
	if (bHotload == true) then		
		Rebuild();
	end
end

-- ===========================================================================
function OnBeginWonderReveal()
	ContextPtr:SetHide( true );
end

-- ===========================================================================
function OnEndWonderReveal()
	ContextPtr:SetHide( false );
end

-- ===========================================================================
function ClearImprovementRecommendations()
	-- Hide previous recommendations
	for i,plotIndex in ipairs(m_RecommendedImprovementPlots) do
		local pRecommendedPlotInstance = GetInstanceAt(plotIndex);
		pRecommendedPlotInstance.ImprovementRecommendationBackground:SetHide(true);
	end

	-- Clear table
	m_RecommendedImprovementPlots = {};
end

-- ===========================================================================
function AddImprovementRecommendationsForCity( pCity:table )
	local pCityAI:table = pCity:GetCityAI();
	if pCityAI then
		local recommendList:table = pCityAI:GetImprovementRecommendations();
		for key,value in pairs(recommendList) do
			local pRecommendedPlotInstance = GetInstanceAt(value.ImprovementLocation);

			-- Get improvement info
			local pImprovementInfo:table = GameInfo.Improvements[value.ImprovementHash];

			-- Update icon
			local textureOffsetX, textureOffsetY, textureSheet = IconManager:FindIconAtlas(pImprovementInfo.Icon, 38);
			pRecommendedPlotInstance.ImprovementRecommendationIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);

			-- Update tooltip
			pRecommendedPlotInstance.ImprovementRecommendationIcon:SetToolTipString(Locale.Lookup("LOC_TOOLTIP_IMPROVEMENT_RECOMMENDATION", pImprovementInfo.Name));

			-- Show recommendation and add to list for clean up later
			pRecommendedPlotInstance.ImprovementRecommendationBackground:SetHide(false);
			table.insert(m_RecommendedImprovementPlots, value.ImprovementLocation);
		end
	end
end

-- ===========================================================================
function ClearSettlementRecommendations()
	-- Hide previous recommendations
	for i,plotIndex in ipairs(m_RecommendedSettlementPlots) do
		local pRecommendedPlotInstance = GetInstanceAt(plotIndex);
		pRecommendedPlotInstance.ImprovementRecommendationBackground:SetHide(true);
	end

	-- Clear table
	m_RecommendedSettlementPlots = {};
end

-- ===========================================================================
function AddSettlementRecommendations()
	local pLocalPlayer:table = Players[Game.GetLocalPlayer()];
	if pLocalPlayer then
		local pGrandAI:table = pLocalPlayer:GetGrandStrategicAI();
		if pGrandAI then
			local pSettlementRecommendations:table = pGrandAI:GetSettlementRecommendations();
			for key,value in pairs(pSettlementRecommendations) do
				local pRecommendedPlotInstance = GetInstanceAt(value.SettlingLocation);

				-- Update icon
				local textureOffsetX, textureOffsetY, textureSheet = IconManager:FindIconAtlas("ICON_UNITOPERATION_FOUND_CITY", 38);
				pRecommendedPlotInstance.ImprovementRecommendationIcon:SetTexture(textureOffsetX, textureOffsetY, textureSheet);

				-- Update tooltip
				pRecommendedPlotInstance.ImprovementRecommendationIcon:SetToolTipString(Locale.Lookup("LOC_TOOLTIP_SETTLEMENT_RECOMMENDATION"));				

				-- Show recommendation and add to list for clean up later
				pRecommendedPlotInstance.ImprovementRecommendationBackground:SetHide(false);
				table.insert(m_RecommendedSettlementPlots, value.SettlingLocation);
			end
		end
	end
end

-- ===========================================================================
function OnUnitSelectionChanged(player, unitId, locationX, locationY, locationZ, isSelected, isEditable)
	ClearImprovementRecommendations();
	ClearSettlementRecommendations();

	-- Are we a builder?
	local pSelectedUnit:table = UI.GetHeadSelectedUnit();
	if pSelectedUnit then 
		if pSelectedUnit:GetBuildCharges() > 0 and GameInfo.Units[pSelectedUnit:GetUnitType()].UnitType == "UNIT_BUILDER" then
			-- If we're within a city then look for any recommended improvements
			local pPlot = Map.GetPlotIndex(pSelectedUnit:GetX(), pSelectedUnit:GetY());
			local pCity:table = Cities.GetPlotPurchaseCity(pPlot);
			if pCity and pCity:GetOwner() == player then
				AddImprovementRecommendationsForCity(pCity);
			end
		elseif GameInfo.Units[pSelectedUnit:GetUnitType()].FoundCity then
			-- Add settlement recommendations if we're a settler
			AddSettlementRecommendations();
		end
	end
end




-- RESOURCE TOOLTIP MOD
-- ===========================================================================
-- Create the resource tooltip string table
function createResourceTooltipString(toolTipItems, pResources, row)
	local resourceAmount:number = pResources:GetResourceAmount(row.Index) or 0;

	-- Simply write the owned amount in parantheses
	toolTipItems[1] = toolTipItems[1] .. " (" .. resourceAmount .. ")";
	return toolTipItems;
end


-- Build the initial resource table
function InitialResourceHandler()
	print("InitialResourceHandler");
	
	local playerID:number         = Game.GetLocalPlayer();
	local player:table            = Players[playerID];
	local pResources:table        = player:GetResources();
	

	-- Build the global owned resource table
	for row in GameInfo.Resources() do
		-- Index
		-- Name
		-- ResourceType
		-- ResourceClassType
		g_ResourcesForPlayer[row.Index] = pResources:GetResourceAmount(row.Index) or 0;
	end


	-- Build the global resource plot table
	for i, val in pairs(g_MapIcons) do
		local pPlot = Map.GetPlotByIndex(val.plotIndex);
		local rType = pPlot:GetResourceType();

		if ( g_ResourcesOnMap[rType] == nil ) then
			g_ResourcesOnMap[rType] = {};
		end
		
		table.insert(g_ResourcesOnMap[rType], val.plotIndex);
	end
end


-- Handler table for the tooltip update
local UpdateResourceTooltipHandler:table = {
	DiplomacyDealEnacted        = function() UpdateResourceTooltip("DiplomacyDealEnacted"); end,
	DiplomacyDealExpired        = function() UpdateResourceTooltip("DiplomacyDealExpired"); end,
	ImprovementActivated        = function() UpdateResourceTooltip("ImprovementActivated"); end,
	ImprovementAddedToMap       = function() UpdateResourceTooltip("ImprovementAddedToMap"); end,
	ImprovementChanged          = function() UpdateResourceTooltip("ImprovementChanged"); end,
	ImprovementOwnershipChanged = function() UpdateResourceTooltip("ImprovementOwnershipChanged"); end,
	ImprovementRemovedFromMap   = function() UpdateResourceTooltip("ImprovementRemovedFromMap"); end,
	ResourceAddedToMap          = function() UpdateResourceTooltip("ResourceAddedToMap"); end,
	ResourceChanged             = function() UpdateResourceTooltip("ResourceChanged"); end,
	ResourceRemovedFromMap      = function() UpdateResourceTooltip("ResourceRemovedFromMap"); end,
};


-- Update the resource tooltip
function UpdateResourceTooltip(identifier)
	print("UpdateResourceTooltip");
	print("identifier: ", identifier);

	local playerID:number         = Game.GetLocalPlayer();
	local player:table            = Players[playerID];
	local pResources:table        = player:GetResources();
	local changeTable:table       = {};

	-- print("g_ResourcesForPlayer: ", g_ResourcesForPlayer);

	-- Compare the amount of available resources to the saved table
	for idx, amount in pairs(g_ResourcesForPlayer) do
		local curAmount:number = pResources:GetResourceAmount(idx) or 0;

		if ( curAmount ~= amount ) then
			print("A resource has changed!");
			print("Resource:   ", GameInfo.Resources[idx].Name);
			print("Old amount: ", amount);
			print("New amount: ", curAmount);
			g_ResourcesForPlayer[idx] = curAmount;
			changeTable[idx]          = curAmount;
		end
	end

	-- Check if we need to update
	if ( table.count(changeTable) < 1 ) then
		-- print("No change necessary");
		return;
	end

	-- Update the tooltip text
	for resourceType, amount in pairs(changeTable) do
		-- Change the tooltip for each changed resource
		for i, plotIndex in pairs(g_ResourcesOnMap[resourceType]) do
			local pInstance = g_MapIcons[plotIndex];
			local tString:string = pInstance.ResourceIcon:GetToolTipString();
			local tooltipStringTable:table = explode("[NEWLINE]", tString);
			
			tooltipStringTable[1] = string.gsub(tooltipStringTable[1], "(%(%d+%))", "(" .. amount .. ")");

			tString = table.concat(tooltipStringTable, "[NEWLINE]");
			pInstance.ResourceIcon:SetToolTipString(tString);
		end
	end
end


-- Explode a string PHP style
-- Source: http://lua-users.org/wiki/MakingLuaLikePhp
-- Credit: http://richard.warburton.it/
function explode(div,str)
	if (div=='') then return false end
	local pos,arr = 0,{}
	for st,sp in function() return string.find(str,div,pos,true) end do
		table.insert(arr,string.sub(str,pos,st-1))
		pos = sp + 1
	end
	table.insert(arr,string.sub(str,pos))
	return arr
end
-- END RESOURCE TOOLTIP MOD



-- ===========================================================================
function Initialize()	
	ContextPtr:SetInitHandler(OnContextInitialize);
	ContextPtr:SetShutdown( OnShutdown );
	
	Events.BeginWonderReveal.Add( OnBeginWonderReveal );
	Events.EndWonderReveal.Add( OnEndWonderReveal );
	Events.LocalPlayerChanged.Add(OnLocalPlayerChanged);
	Events.ResourceVisibilityChanged.Add(OnResourceVisibilityChanged);
	Events.ResourceAddedToMap.Add(OnResourceChanged);
	Events.ResourceRemovedFromMap.Add(OnResourceRemovedFromMap);
	Events.PlotVisibilityChanged.Add(OnPlotVisibilityChanged);
	Events.PlotMarkerChanged.Add(OnPlotMarkersChanged);
	Events.UnitSelectionChanged.Add( OnUnitSelectionChanged );
	Events.UserOptionChanged.Add(OnUserOptionChanged);
	Events.ResearchCompleted.Add(OnResearchCompleted);
	Events.CivicCompleted.Add(OnCivicCompleted);

	-- RESOURCE TOOLTIP MOD
	-- Build the initial resource table
	Events.ResourceAddedToMap.Add(          InitialResourceHandler );

	-- Observe events that could possibly change the amount of resources
	Events.DiplomacyDealEnacted.Add(        UpdateResourceTooltipHandler["DiplomacyDealEnacted"] );
	Events.DiplomacyDealExpired.Add(        UpdateResourceTooltipHandler["DiplomacyDealExpired"] );
	Events.ImprovementActivated.Add(        UpdateResourceTooltipHandler["ImprovementActivated"] );
	Events.ImprovementAddedToMap.Add(       UpdateResourceTooltipHandler["ImprovementAddedToMap"] );
	Events.ImprovementChanged.Add(          UpdateResourceTooltipHandler["ImprovementChanged"] );
	Events.ImprovementOwnershipChanged.Add( UpdateResourceTooltipHandler["ImprovementOwnershipChanged"] );
	Events.ImprovementRemovedFromMap.Add(   UpdateResourceTooltipHandler["ImprovementRemovedFromMap"] );
	Events.ResourceAddedToMap.Add(          UpdateResourceTooltipHandler["ResourceAddedToMap"] );
	Events.ResourceChanged.Add(             UpdateResourceTooltipHandler["ResourceChanged"] );
	Events.ResourceRemovedFromMap.Add(      UpdateResourceTooltipHandler["ResourceRemovedFromMap"] );
	-- END RESOURCE TOOLTIP MOD


	for row in GameInfo.Resources() do
		if row.PrereqTech ~= nil then
			table.insert(m_techsThatUnlockResources, row.PrereqTech);
		end
		if row.PrereqCivic~= nil then
			table.insert(m_civicsThatUnlockResources, row.PrereqCivic);
		end
	end

	for row in GameInfo.Improvements() do
		if (row.PrereqTech ~= nil) then
			table.insert(m_techsThatUnlockImprovements, row.PrereqTech);
		end
	end
	
end
Initialize();

